@@ -4,8 +4,6 @@
 # w/ additions by Travis Oliphant, March 2002
 #              and Jake Vanderplas, August 2012
 
-from __future__ import division, print_function, absolute_import
-
 from warnings import warn
 import numpy as np
 from numpy import atleast_1d, atleast_2d
@@ -18,7 +16,7 @@ from ._solve_toeplitz import levinson
 
 __all__ = ['solve', 'solve_triangular', 'solveh_banded', 'solve_banded',
            'solve_toeplitz', 'solve_circulant', 'inv', 'det', 'lstsq',
-           'pinv', 'pinv2', 'pinvh', 'matrix_balance']
+           'pinv', 'pinv2', 'pinvh', 'matrix_balance', 'matmul_toeplitz']
 
 
 # Linear equations
@@ -124,16 +122,16 @@ def solve(a, b, sym_pos=False, lower=False, overwrite_a=False,
 
     Notes
     -----
-    If the input b matrix is a 1D array with N elements, when supplied
+    If the input b matrix is a 1-D array with N elements, when supplied
     together with an NxN input a, it is assumed as a valid column vector
     despite the apparent size mismatch. This is compatible with the
-    numpy.dot() behavior and the returned result is still 1D array.
+    numpy.dot() behavior and the returned result is still 1-D array.
 
-    The generic, symmetric, hermitian and positive definite solutions are
+    The generic, symmetric, Hermitian and positive definite solutions are
     obtained via calling ?GESV, ?SYSV, ?HESV, and ?POSV routines of
     LAPACK respectively.
     """
-    # Flags for 1D or nD right hand side
+    # Flags for 1-D or N-D right-hand side
     b_is_1D = False
 
     a1 = atleast_2d(_asarray_validated(a, check_finite=check_finite))
@@ -147,7 +145,7 @@ def solve(a, b, sym_pos=False, lower=False, overwrite_a=False,
         raise ValueError('Input a needs to be a square matrix.')
 
     if n != b1.shape[0]:
-        # Last chance to catch 1x1 scalar a and 1D b arrays
+        # Last chance to catch 1x1 scalar a and 1-D b arrays
         if not (n == 1 and b1.size != 0):
             raise ValueError('Input b has to have same number of rows as '
                              'input a')
@@ -156,7 +154,7 @@ def solve(a, b, sym_pos=False, lower=False, overwrite_a=False,
     if b1.size == 0:
         return np.asfortranarray(b1.copy())
 
-    # regularize 1D b arrays to 2D
+    # regularize 1-D b arrays to 2D
     if b1.ndim == 1:
         if n == 1:
             b1 = b1[None, :]
@@ -337,21 +335,27 @@ def solve_triangular(a, b, trans=0, lower=False, unit_diagonal=False,
     if len(a1.shape) != 2 or a1.shape[0] != a1.shape[1]:
         raise ValueError('expected square matrix')
     if a1.shape[0] != b1.shape[0]:
-        raise ValueError('incompatible dimensions')
+        raise ValueError('shapes of a {} and b {} are incompatible'
+                         .format(a1.shape, b1.shape))
     overwrite_b = overwrite_b or _datacopied(b1, b)
     if debug:
         print('solve:overwrite_b=', overwrite_b)
     trans = {'N': 0, 'T': 1, 'C': 2}.get(trans, trans)
     trtrs, = get_lapack_funcs(('trtrs',), (a1, b1))
-    x, info = trtrs(a1, b1, overwrite_b=overwrite_b, lower=lower,
-                    trans=trans, unitdiag=unit_diagonal)
+    if a1.flags.f_contiguous or trans == 2:
+        x, info = trtrs(a1, b1, overwrite_b=overwrite_b, lower=lower,
+                        trans=trans, unitdiag=unit_diagonal)
+    else:
+        # transposed system is solved since trtrs expects Fortran ordering
+        x, info = trtrs(a1.T, b1, overwrite_b=overwrite_b, lower=not lower,
+                        trans=not trans, unitdiag=unit_diagonal)
 
     if info == 0:
         return x
     if info > 0:
         raise LinAlgError("singular matrix: resolution failed at diagonal %d" %
                           (info-1))
-    raise ValueError('illegal value in %d-th argument of internal trtrs' %
+    raise ValueError('illegal value in %dth argument of internal trtrs' %
                      (-info))
 
 
@@ -391,7 +395,7 @@ def solve_banded(l_and_u, ab, b, overwrite_ab=False, overwrite_b=False,
     Returns
     -------
     x : (M,) or (M, K) ndarray
-        The solution to the system a x = b.  Returned shape depends on the
+        The solution to the system a x = b. Returned shape depends on the
         shape of `b`.
 
     Examples
@@ -405,7 +409,7 @@ def solve_banded(l_and_u, ab, b, overwrite_ab=False, overwrite_b=False,
             [0  0  0  1  1]       [3]
 
     There is one nonzero diagonal below the main diagonal (l = 1), and
-    two above (u = 2).  The diagonal banded form of the matrix is::
+    two above (u = 2). The diagonal banded form of the matrix is::
 
              [*  * -1 -1 -1]
         ab = [*  2  2  2  2]
@@ -513,7 +517,7 @@ def solveh_banded(ab, b, overwrite_ab=False, overwrite_b=False, lower=False,
     Returns
     -------
     x : (M,) or (M, K) ndarray
-        The solution to the system a x = b.  Shape of return matches shape
+        The solution to the system a x = b. Shape of return matches shape
         of `b`.
 
     Examples
@@ -530,7 +534,7 @@ def solveh_banded(ab, b, overwrite_ab=False, overwrite_b=False, lower=False,
     >>> from scipy.linalg import solveh_banded
 
     `ab` contains the main diagonal and the nonzero diagonals below the
-    main diagonal.  That is, we use the lower form:
+    main diagonal. That is, we use the lower form:
 
     >>> ab = np.array([[ 4,  5,  6,  7, 8, 9],
     ...                [ 2,  2,  2,  2, 2, 0],
@@ -584,9 +588,9 @@ def solveh_banded(ab, b, overwrite_ab=False, overwrite_b=False, lower=False,
         c, x, info = pbsv(a1, b1, lower=lower, overwrite_ab=overwrite_ab,
                           overwrite_b=overwrite_b)
     if info > 0:
-        raise LinAlgError("%d-th leading minor not positive definite" % info)
+        raise LinAlgError("%dth leading minor not positive definite" % info)
     if info < 0:
-        raise ValueError('illegal value in %d-th argument of internal '
+        raise ValueError('illegal value in %dth argument of internal '
                          'pbsv' % -info)
     return x
 
@@ -595,7 +599,7 @@ def solve_toeplitz(c_or_cr, b, check_finite=True):
     """Solve a Toeplitz system using Levinson Recursion
 
     The Toeplitz matrix has constant diagonals, with c as its first column
-    and r as its first row.  If r is not given, ``r == conjugate(c)`` is
+    and r as its first row. If r is not given, ``r == conjugate(c)`` is
     assumed.
 
     Parameters
@@ -605,7 +609,7 @@ def solve_toeplitz(c_or_cr, b, check_finite=True):
         actual shape of ``c``, it will be converted to a 1-D array. If not
         supplied, ``r = conjugate(c)`` is assumed; in this case, if c[0] is
         real, the Toeplitz matrix is Hermitian. r[0] is ignored; the first row
-        of the Toeplitz matrix is ``[c[0], r[1:]]``.  Whatever the actual shape
+        of the Toeplitz matrix is ``[c[0], r[1:]]``. Whatever the actual shape
         of ``r``, it will be converted to a 1-D array.
     b : (M,) or (M, K) array_like
         Right-hand side in ``T x = b``.
@@ -617,7 +621,7 @@ def solve_toeplitz(c_or_cr, b, check_finite=True):
     Returns
     -------
     x : (M,) or (M, K) ndarray
-        The solution to the system ``T x = b``.  Shape of return matches shape
+        The solution to the system ``T x = b``. Shape of return matches shape
         of `b`.
 
     See Also
@@ -661,35 +665,19 @@ def solve_toeplitz(c_or_cr, b, check_finite=True):
     # If numerical stability of this algorithm is a problem, a future
     # developer might consider implementing other O(N^2) Toeplitz solvers,
     # such as GKO (https://www.jstor.org/stable/2153371) or Bareiss.
-    if isinstance(c_or_cr, tuple):
-        c, r = c_or_cr
-        c = _asarray_validated(c, check_finite=check_finite).ravel()
-        r = _asarray_validated(r, check_finite=check_finite).ravel()
-    else:
-        c = _asarray_validated(c_or_cr, check_finite=check_finite).ravel()
-        r = c.conjugate()
 
-    # Form a 1D array of values to be used in the matrix, containing a reversed
-    # copy of r[1:], followed by c.
+    r, c, b, dtype, b_shape = _validate_args_for_toeplitz_ops(
+        c_or_cr, b, check_finite, keep_b_shape=True)
+
+    # Form a 1-D array of values to be used in the matrix, containing a
+    # reversed copy of r[1:], followed by c.
     vals = np.concatenate((r[-1:0:-1], c))
     if b is None:
         raise ValueError('illegal value, `b` is a required argument')
 
-    b = _asarray_validated(b)
-    if vals.shape[0] != (2*b.shape[0] - 1):
-        raise ValueError('incompatible dimensions')
-    if np.iscomplexobj(vals) or np.iscomplexobj(b):
-        vals = np.asarray(vals, dtype=np.complex128, order='c')
-        b = np.asarray(b, dtype=np.complex128)
-    else:
-        vals = np.asarray(vals, dtype=np.double, order='c')
-        b = np.asarray(b, dtype=np.double)
-
     if b.ndim == 1:
         x, _ = levinson(vals, np.ascontiguousarray(b))
     else:
-        b_shape = b.shape
-        b = b.reshape(b.shape[0], -1)
         x = np.column_stack([levinson(vals, np.ascontiguousarray(b[:, i]))[0]
                              for i in range(b.shape[1])])
         x = x.reshape(*b_shape)
@@ -712,13 +700,13 @@ def solve_circulant(c, b, singular='raise', tol=None,
 
     `C` is the circulant matrix associated with the vector `c`.
 
-    The system is solved by doing division in Fourier space.  The
+    The system is solved by doing division in Fourier space. The
     calculation is::
 
         x = ifft(fft(b) / fft(c))
 
     where `fft` and `ifft` are the fast Fourier transform and its inverse,
-    respectively.  For a large vector `c`, this is *much* faster than
+    respectively. For a large vector `c`, this is *much* faster than
     solving the system with the full circulant matrix.
 
     Parameters
@@ -730,12 +718,12 @@ def solve_circulant(c, b, singular='raise', tol=None,
     singular : str, optional
         This argument controls how a near singular circulant matrix is
         handled.  If `singular` is "raise" and the circulant matrix is
-        near singular, a `LinAlgError` is raised.  If `singular` is
-        "lstsq", the least squares solution is returned.  Default is "raise".
+        near singular, a `LinAlgError` is raised. If `singular` is
+        "lstsq", the least squares solution is returned. Default is "raise".
     tol : float, optional
         If any eigenvalue of the circulant matrix has an absolute value
         that is less than or equal to `tol`, the matrix is considered to be
-        near singular.  If not given, `tol` is set to::
+        near singular. If not given, `tol` is set to::
 
             tol = abs_eigs.max() * abs_eigs.size * np.finfo(np.float64).eps
 
@@ -743,15 +731,15 @@ def solve_circulant(c, b, singular='raise', tol=None,
         of the circulant matrix.
     caxis : int
         When `c` has dimension greater than 1, it is viewed as a collection
-        of circulant vectors.  In this case, `caxis` is the axis of `c` that
+        of circulant vectors. In this case, `caxis` is the axis of `c` that
         holds the vectors of circulant coefficients.
     baxis : int
         When `b` has dimension greater than 1, it is viewed as a collection
-        of vectors.  In this case, `baxis` is the axis of `b` that holds the
+        of vectors. In this case, `baxis` is the axis of `b` that holds the
         right-hand side vectors.
     outaxis : int
         When `c` or `b` are multidimensional, the value returned by
-        `solve_circulant` is multidimensional.  In this case, `outaxis` is
+        `solve_circulant` is multidimensional. In this case, `outaxis` is
         the axis of the result that holds the solution vectors.
 
     Returns
@@ -770,7 +758,7 @@ def solve_circulant(c, b, singular='raise', tol=None,
 
     Notes
     -----
-    For a one-dimensional vector `c` with length `m`, and an array `b`
+    For a 1-D vector `c` with length `m`, and an array `b`
     with shape ``(m, ...)``,
 
         solve_circulant(c, b)
@@ -824,15 +812,15 @@ def solve_circulant(c, b, singular='raise', tol=None,
     >>> b = np.arange(15).reshape(-1, 5)
 
     We want to solve all combinations of circulant matrices and `b` vectors,
-    with the result stored in an array with shape (2, 3, 5).  When we
+    with the result stored in an array with shape (2, 3, 5). When we
     disregard the axes of `c` and `b` that hold the vectors of coefficients,
     the shapes of the collections are (2,) and (3,), respectively, which are
-    not compatible for broadcasting.  To have a broadcast result with shape
+    not compatible for broadcasting. To have a broadcast result with shape
     (2, 3), we add a trivial dimension to `c`: ``c[:, np.newaxis, :]`` has
-    shape (2, 1, 5).  The last dimension holds the coefficients of the
+    shape (2, 1, 5). The last dimension holds the coefficients of the
     circulant matrices, so when we call `solve_circulant`, we can use the
-    default ``caxis=-1``.  The coefficients of the `b` vectors are in the last
-    dimension of the array `b`, so we use ``baxis=-1``.  If we use the
+    default ``caxis=-1``. The coefficients of the `b` vectors are in the last
+    dimension of the array `b`, so we use ``baxis=-1``. If we use the
     default `outaxis`, the result will have shape (5, 2, 3), so we'll use
     ``outaxis=-1`` to put the solution vectors in the last dimension.
 
@@ -859,7 +847,8 @@ def solve_circulant(c, b, singular='raise', tol=None,
     b = np.atleast_1d(b)
     nb = _get_axis_len("b", b, baxis)
     if nc != nb:
-        raise ValueError('Incompatible c and b axis lengths')
+        raise ValueError('Shapes of c {} and b {} are incompatible'
+                         .format(c.shape, b.shape))
 
     fc = np.fft.fft(np.rollaxis(c, caxis, c.ndim), axis=-1)
     abs_fc = np.abs(fc)
@@ -887,7 +876,7 @@ def solve_circulant(c, b, singular='raise', tol=None,
 
     if is_near_singular:
         # `near_zeros` is a boolean array, same shape as `c`, that is
-        # True where `fc` is (near) zero.  `q` is the broadcasted result
+        # True where `fc` is (near) zero. `q` is the broadcasted result
         # of fb / fc, so to set the values of `q` to 0 where `fc` is near
         # zero, we use a mask that is the broadcast result of an array
         # of True values shaped like `b` with `near_zeros`.
@@ -928,7 +917,7 @@ def inv(a, overwrite_a=False, check_finite=True):
     LinAlgError
         If `a` is singular.
     ValueError
-        If `a` is not square, or not 2-dimensional.
+        If `a` is not square, or not 2D.
 
     Examples
     --------
@@ -1049,7 +1038,7 @@ def lstsq(a, b, cond=None, overwrite_a=False, overwrite_b=False,
     Parameters
     ----------
     a : (M, N) array_like
-        Left hand side array
+        Left-hand side array
     b : (M,) or (M, K) array_like
         Right hand side array
     cond : float, optional
@@ -1152,7 +1141,7 @@ def lstsq(a, b, cond=None, overwrite_a=False, overwrite_b=False,
     a1 = _asarray_validated(a, check_finite=check_finite)
     b1 = _asarray_validated(b, check_finite=check_finite)
     if len(a1.shape) != 2:
-        raise ValueError('Input array a should be 2-D')
+        raise ValueError('Input array a should be 2D')
     m, n = a1.shape
     if len(b1.shape) == 2:
         nrhs = b1.shape[1]
@@ -1256,7 +1245,7 @@ def pinv(a, cond=None, rcond=None, return_rank=False, check_finite=True):
     a : (M, N) array_like
         Matrix to be pseudo-inverted.
     cond, rcond : float, optional
-        Cutoff factor for 'small' singular values. In `lstsq`, 
+        Cutoff factor for 'small' singular values. In `lstsq`,
         singular values less than ``cond*largest_singular_value`` will be
         considered as zero. If both are omitted, the default value
         ``max(M, N) * eps`` is passed to `lstsq` where ``eps`` is the
@@ -1278,7 +1267,7 @@ def pinv(a, cond=None, rcond=None, return_rank=False, check_finite=True):
     B : (N, M) ndarray
         The pseudo-inverse of matrix `a`.
     rank : int
-        The effective rank of the matrix.  Returned if return_rank == True
+        The effective rank of the matrix. Returned if return_rank == True
 
     Raises
     ------
@@ -1297,7 +1286,9 @@ def pinv(a, cond=None, rcond=None, return_rank=False, check_finite=True):
 
     """
     a = _asarray_validated(a, check_finite=check_finite)
-    b = np.identity(a.shape[0], dtype=a.dtype)
+    # If a is sufficiently tall it is cheaper to compute using the transpose
+    trans = a.shape[0] / a.shape[1] >= 1.1
+    b = np.eye(a.shape[1] if trans else a.shape[0], dtype=a.dtype)
 
     if rcond is not None:
         cond = rcond
@@ -1305,12 +1296,13 @@ def pinv(a, cond=None, rcond=None, return_rank=False, check_finite=True):
     if cond is None:
         cond = max(a.shape) * np.spacing(a.real.dtype.type(1))
 
-    x, resids, rank, s = lstsq(a, b, cond=cond, check_finite=False)
+    x, resids, rank, s = lstsq(a.T if trans else a, b,
+                               cond=cond, check_finite=False)
 
     if return_rank:
-        return x, rank
+        return (x.T if trans else x), rank
     else:
-        return x
+        return x.T if trans else x
 
 
 def pinv2(a, cond=None, rcond=None, return_rank=False, check_finite=True):
@@ -1347,7 +1339,7 @@ def pinv2(a, cond=None, rcond=None, return_rank=False, check_finite=True):
     B : (N, M) ndarray
         The pseudo-inverse of matrix `a`.
     rank : int
-        The effective rank of the matrix.  Returned if `return_rank` is True.
+        The effective rank of the matrix. Returned if `return_rank` is True.
 
     Raises
     ------
@@ -1499,7 +1491,7 @@ def matrix_balance(A, permute=True, scale=True, separate=False,
         will not be scaled.
     separate : bool, optional
         This switches from returning a full matrix of the transformation
-        to a tuple of two separate 1D permutation and scaling arrays.
+        to a tuple of two separate 1-D permutation and scaling arrays.
     overwrite_a : bool, optional
         This is passed to xGEBAL directly. Essentially, overwrites the result
         to the data. It might increase the space efficiency. See LAPACK manual
@@ -1556,11 +1548,10 @@ def matrix_balance(A, permute=True, scale=True, separate=False,
     ----------
     .. [1] : B.N. Parlett and C. Reinsch, "Balancing a Matrix for
        Calculation of Eigenvalues and Eigenvectors", Numerische Mathematik,
-       Vol.13(4), 1969, DOI:10.1007/BF02165404
+       Vol.13(4), 1969, :doi:`10.1007/BF02165404`
 
     .. [2] : R. James, J. Langou, B.R. Lowery, "On matrix balancing and
-       eigenvector computation", 2014, Available online:
-       https://arxiv.org/abs/1401.5766
+       eigenvector computation", 2014, :arxiv:`1401.5766`
 
     .. [3] :  D.S. Watkins. A case where balancing is harmful.
        Electron. Trans. Numer. Anal, Vol.23, 2006.
@@ -1612,3 +1603,229 @@ def matrix_balance(A, permute=True, scale=True, separate=False,
     iperm[perm] = np.arange(n)
 
     return B, np.diag(scaling)[iperm, :]
+
+
+def _validate_args_for_toeplitz_ops(c_or_cr, b, check_finite, keep_b_shape,
+                                    enforce_square=True):
+    """Validate arguments and format inputs for toeplitz functions
+
+    Parameters
+    ----------
+    c_or_cr : array_like or tuple of (array_like, array_like)
+        The vector ``c``, or a tuple of arrays (``c``, ``r``). Whatever the
+        actual shape of ``c``, it will be converted to a 1-D array. If not
+        supplied, ``r = conjugate(c)`` is assumed; in this case, if c[0] is
+        real, the Toeplitz matrix is Hermitian. r[0] is ignored; the first row
+        of the Toeplitz matrix is ``[c[0], r[1:]]``. Whatever the actual shape
+        of ``r``, it will be converted to a 1-D array.
+    b : (M,) or (M, K) array_like
+        Right-hand side in ``T x = b``.
+    check_finite : bool
+        Whether to check that the input matrices contain only finite numbers.
+        Disabling may give a performance gain, but may result in problems
+        (result entirely NaNs) if the inputs do contain infinities or NaNs.
+    keep_b_shape: bool
+        Whether to convert a (M,) dimensional b into a (M, 1) dimensional
+        matrix.
+    enforce_square: bool, optional
+        If True (default), this verifies that the Toeplitz matrix is square.
+
+    Returns
+    -------
+    r : array
+        1d array corresponding to the first row of the Toeplitz matrix.
+    c: array
+        1d array corresponding to the first column of the Toeplitz matrix.
+    b: array
+        (M,), (M, 1) or (M, K) dimensional array, post validation,
+        corresponding to ``b``.
+    dtype: numpy datatype
+        ``dtype`` stores the datatype of ``r``, ``c`` and ``b``. If any of
+        ``r``, ``c`` or ``b`` are complex, ``dtype`` is ``np.complex128``,
+        otherwise, it is ``np.float``.
+    b_shape: tuple
+        Shape of ``b`` after passing it through ``_asarray_validated``.
+
+    """
+
+    if isinstance(c_or_cr, tuple):
+        c, r = c_or_cr
+        c = _asarray_validated(c, check_finite=check_finite).ravel()
+        r = _asarray_validated(r, check_finite=check_finite).ravel()
+    else:
+        c = _asarray_validated(c_or_cr, check_finite=check_finite).ravel()
+        r = c.conjugate()
+
+    if b is None:
+        raise ValueError('`b` must be an array, not None.')
+
+    b = _asarray_validated(b, check_finite=check_finite)
+    b_shape = b.shape
+
+    is_not_square = r.shape[0] != c.shape[0]
+    if (enforce_square and is_not_square) or b.shape[0] != r.shape[0]:
+        raise ValueError('Incompatible dimensions.')
+
+    is_cmplx = np.iscomplexobj(r) or np.iscomplexobj(c) or np.iscomplexobj(b)
+    dtype = np.complex128 if is_cmplx else np.double
+    r, c, b = (np.asarray(i, dtype=dtype) for i in (r, c, b))
+
+    if b.ndim == 1 and not keep_b_shape:
+        b = b.reshape(-1, 1)
+    elif b.ndim != 1:
+        b = b.reshape(b.shape[0], -1)
+
+    return r, c, b, dtype, b_shape
+
+
+def matmul_toeplitz(c_or_cr, x, check_finite=False, workers=None):
+    """Efficient Toeplitz Matrix-Matrix Multiplication using FFT
+
+    This function returns the matrix multiplication between a Toeplitz
+    matrix and a dense matrix.
+
+    The Toeplitz matrix has constant diagonals, with c as its first column
+    and r as its first row. If r is not given, ``r == conjugate(c)`` is
+    assumed.
+
+    Parameters
+    ----------
+    c_or_cr : array_like or tuple of (array_like, array_like)
+        The vector ``c``, or a tuple of arrays (``c``, ``r``). Whatever the
+        actual shape of ``c``, it will be converted to a 1-D array. If not
+        supplied, ``r = conjugate(c)`` is assumed; in this case, if c[0] is
+        real, the Toeplitz matrix is Hermitian. r[0] is ignored; the first row
+        of the Toeplitz matrix is ``[c[0], r[1:]]``. Whatever the actual shape
+        of ``r``, it will be converted to a 1-D array.
+    x : (M,) or (M, K) array_like
+        Matrix with which to multiply.
+    check_finite : bool, optional
+        Whether to check that the input matrices contain only finite numbers.
+        Disabling may give a performance gain, but may result in problems
+        (result entirely NaNs) if the inputs do contain infinities or NaNs.
+    workers : int, optional
+        To pass to scipy.fft.fft and ifft. Maximum number of workers to use
+        for parallel computation. If negative, the value wraps around from
+        ``os.cpu_count()``. See scipy.fft.fft for more details.
+
+    Returns
+    -------
+    T @ x : (M,) or (M, K) ndarray
+        The result of the matrix multiplication ``T @ x``. Shape of return
+        matches shape of `x`.
+
+    See Also
+    --------
+    toeplitz : Toeplitz matrix
+    solve_toeplitz : Solve a Toeplitz system using Levinson Recursion
+
+    Notes
+    -----
+    The Toeplitz matrix is embedded in a circulant matrix and the FFT is used
+    to efficiently calculate the matrix-matrix product.
+
+    Because the computation is based on the FFT, integer inputs will
+    result in floating point outputs.  This is unlike NumPy's `matmul`,
+    which preserves the data type of the input.
+
+    This is partly based on the implementation that can be found in [1]_,
+    licensed under the MIT license. More information about the method can be
+    found in reference [2]_. References [3]_ and [4]_ have more reference
+    implementations in Python.
+
+    .. versionadded:: 1.6.0
+
+    References
+    ----------
+    .. [1] Jacob R Gardner, Geoff Pleiss, David Bindel, Kilian
+       Q Weinberger, Andrew Gordon Wilson, "GPyTorch: Blackbox Matrix-Matrix
+       Gaussian Process Inference with GPU Acceleration" with contributions
+       from Max Balandat and Ruihan Wu. Available online:
+       https://github.com/cornellius-gp/gpytorch
+
+    .. [2] J. Demmel, P. Koev, and X. Li, "A Brief Survey of Direct Linear
+       Solvers". In Z. Bai, J. Demmel, J. Dongarra, A. Ruhe, and H. van der
+       Vorst, editors. Templates for the Solution of Algebraic Eigenvalue
+       Problems: A Practical Guide. SIAM, Philadelphia, 2000. Available at:
+       http://www.netlib.org/utk/people/JackDongarra/etemplates/node384.html
+
+    .. [3] R. Scheibler, E. Bezzam, I. Dokmanic, Pyroomacoustics: A Python
+       package for audio room simulations and array processing algorithms,
+       Proc. IEEE ICASSP, Calgary, CA, 2018.
+       https://github.com/LCAV/pyroomacoustics/blob/pypi-release/
+       pyroomacoustics/adaptive/util.py
+
+    .. [4] Marano S, Edwards B, Ferrari G and Fah D (2017), "Fitting
+       Earthquake Spectra: Colored Noise and Incomplete Data", Bulletin of
+       the Seismological Society of America., January, 2017. Vol. 107(1),
+       pp. 276-291.
+
+    Examples
+    --------
+    Multiply the Toeplitz matrix T with matrix x::
+
+            [ 1 -1 -2 -3]       [1 10]
+        T = [ 3  1 -1 -2]   x = [2 11]
+            [ 6  3  1 -1]       [2 11]
+            [10  6  3  1]       [5 19]
+
+    To specify the Toeplitz matrix, only the first column and the first
+    row are needed.
+
+    >>> c = np.array([1, 3, 6, 10])    # First column of T
+    >>> r = np.array([1, -1, -2, -3])  # First row of T
+    >>> x = np.array([[1, 10], [2, 11], [2, 11], [5, 19]])
+
+    >>> from scipy.linalg import toeplitz, matmul_toeplitz
+    >>> matmul_toeplitz((c, r), x)
+    array([[-20., -80.],
+           [ -7.,  -8.],
+           [  9.,  85.],
+           [ 33., 218.]])
+
+    Check the result by creating the full Toeplitz matrix and
+    multiplying it by ``x``.
+
+    >>> toeplitz(c, r) @ x
+    array([[-20, -80],
+           [ -7,  -8],
+           [  9,  85],
+           [ 33, 218]])
+
+    The full matrix is never formed explicitly, so this routine
+    is suitable for very large Toeplitz matrices.
+
+    >>> n = 1000000
+    >>> matmul_toeplitz([1] + [0]*(n-1), np.ones(n))
+    array([1., 1., 1., ..., 1., 1., 1.])
+
+    """
+
+    from ..fft import fft, ifft, rfft, irfft
+
+    r, c, x, dtype, x_shape = _validate_args_for_toeplitz_ops(
+        c_or_cr, x, check_finite, keep_b_shape=False, enforce_square=False)
+    n, m = x.shape
+
+    T_nrows = len(c)
+    T_ncols = len(r)
+    p = T_nrows + T_ncols - 1  # equivalent to len(embedded_col)
+
+    embedded_col = np.concatenate((c, r[-1:0:-1]))
+
+    if np.iscomplexobj(embedded_col) or np.iscomplexobj(x):
+        fft_mat = fft(embedded_col, axis=0, workers=workers).reshape(-1, 1)
+        fft_x = fft(x, n=p, axis=0, workers=workers)
+
+        mat_times_x = ifft(fft_mat*fft_x, axis=0,
+                           workers=workers)[:T_nrows, :]
+    else:
+        # Real inputs; using rfft is faster
+        fft_mat = rfft(embedded_col, axis=0, workers=workers).reshape(-1, 1)
+        fft_x = rfft(x, n=p, axis=0, workers=workers)
+
+        mat_times_x = irfft(fft_mat*fft_x, axis=0,
+                            workers=workers, n=p)[:T_nrows, :]
+
+    return_shape = (T_nrows,) if len(x_shape) == 1 else (T_nrows, m)
+    return mat_times_x.reshape(*return_shape)
