@@ -4,20 +4,20 @@ Python wrappers for Orthogonal Distance Regression (ODRPACK).
 Notes
 =====
 
-* Array formats -- FORTRAN stores its arrays in memory column first, i.e. an
+* Array formats -- FORTRAN stores its arrays in memory column first, i.e., an
   array element A(i, j, k) will be next to A(i+1, j, k). In C and, consequently,
   NumPy, arrays are stored row first: A[i, j, k] is next to A[i, j, k+1]. For
   efficiency and convenience, the input and output arrays of the fitting
   function (and its Jacobians) are passed to FORTRAN without transposition.
   Therefore, where the ODRPACK documentation says that the X array is of shape
   (N, M), it will be passed to the Python function as an array of shape (M, N).
-  If M==1, the one-dimensional case, then nothing matters; if M>1, then your
+  If M==1, the 1-D case, then nothing matters; if M>1, then your
   Python functions will be dealing with arrays that are indexed in reverse of
-  the ODRPACK documentation. No real biggie, but watch out for your indexing of
-  the Jacobians: the i,j'th elements (@f_i/@x_j) evaluated at the n'th
+  the ODRPACK documentation. No real issue, but watch out for your indexing of
+  the Jacobians: the i,jth elements (@f_i/@x_j) evaluated at the nth
   observation will be returned as jacd[j, i, n]. Except for the Jacobians, it
   really is easier to deal with x[0] and x[1] than x[:,0] and x[:,1]. Of course,
-  you can always use the transpose() function from scipy explicitly.
+  you can always use the transpose() function from SciPy explicitly.
 
 * Examples -- See the accompanying file test/test.py for examples of how to set
   up fits of your own. Some are taken from the User's Guide; some are from
@@ -35,8 +35,7 @@ Robert Kern
 robert.kern@gmail.com
 
 """
-
-from __future__ import division, print_function, absolute_import
+import os
 
 import numpy
 from warnings import warn
@@ -62,7 +61,7 @@ class OdrError(Exception):
     """
     Exception indicating an error in fitting.
 
-    This is raised by `scipy.odr` if an error occurs during fitting.
+    This is raised by `~scipy.odr.odr` if an error occurs during fitting.
     """
     pass
 
@@ -72,10 +71,11 @@ class OdrStop(Exception):
     Exception stopping fitting.
 
     You can raise this exception in your objective function to tell
-    `scipy.odr` to stop fitting.
+    `~scipy.odr.odr` to stop fitting.
     """
     pass
 
+
 # Backwards compatibility
 odr_error = OdrError
 odr_stop = OdrStop
@@ -217,12 +217,12 @@ class Data(object):
         variable), then this vector is the diagonal of the covariant weighting
         matrix for all data points.
         If `wd` is a rank-1 array of length n (the number of data points), then
-        the i'th element is the weight for the i'th input variable observation
+        the i'th element is the weight for the ith input variable observation
         (single-dimensional only).
         If `wd` is a rank-2 array of shape (m, m), then this is the full
         covariant weighting matrix broadcast to each observation.
         If `wd` is a rank-2 array of shape (m, n), then `wd[:,i]` is the
-        diagonal of the covariant weighting matrix for the i'th observation.
+        diagonal of the covariant weighting matrix for the ith observation.
         If `wd` is a rank-3 array of shape (m, m, n), then `wd[:,:,i]` is the
         full specification of the covariant weighting matrix for each
         observation.
@@ -245,7 +245,7 @@ class Data(object):
     specified value of `y`.
 
     The `we` argument weights the effect a deviation in the response variable
-    has on the fit.  The `wd` argument weights the effect a deviation in the
+    has on the fit. The `wd` argument weights the effect a deviation in the
     input variable has on the fit. To handle multidimensional inputs and
     responses easily, the structure of these arguments has the n'th
     dimensional axis first. These arguments heavily use the structured
@@ -343,14 +343,14 @@ class RealData(Data):
 
     `covx` and `covy` are arrays of covariance matrices and are converted to
     weights by performing a matrix inversion on each observation's covariance
-    matrix.  For example, ``we[i] = numpy.linalg.inv(covy[i])``.
+    matrix. For example, ``we[i] = numpy.linalg.inv(covy[i])``.
 
     These arguments follow the same structured argument conventions as wd and
     we only restricted by their natures: `sx` and `sy` can't be rank-3, but
     `covx` and `covy` can be.
 
     Only set *either* `sx` or `covx` (not both). Setting both will raise an
-    exception.  Same with `sy` and `covy`.
+    exception. Same with `sy` and `covy`.
 
     """
 
@@ -397,7 +397,7 @@ class RealData(Data):
         """ Convert covariance matrix(-ices) to weights.
         """
 
-        from numpy.dual import inv
+        from scipy.linalg import inv
 
         if len(cov.shape) == 2:
             return inv(cov)
@@ -477,12 +477,12 @@ class Model(object):
 
     `x`
         if the input data is single-dimensional, then `x` is rank-1
-        array; i.e. ``x = array([1, 2, 3, ...]); x.shape = (n,)``
+        array; i.e., ``x = array([1, 2, 3, ...]); x.shape = (n,)``
         If the input data is multi-dimensional, then `x` is a rank-2 array;
         i.e., ``x = array([[1, 2, ...], [2, 4, ...]]); x.shape = (m, n)``.
         In all cases, it has the same shape as the input data array passed to
-        `odr`. `m` is the dimensionality of the input data, `n` is the number
-        of observations.
+        `~scipy.odr.odr`. `m` is the dimensionality of the input data,
+        `n` is the number of observations.
     `y`
         if the response variable is single-dimensional, then `y` is a
         rank-1 array, i.e., ``y = array([2, 4, ...]); y.shape = (n,)``.
@@ -495,11 +495,11 @@ class Model(object):
     `fjacb`
         if the response variable is multi-dimensional, then the
         return array's shape is `(q, p, n)` such that ``fjacb(x,beta)[l,k,i] =
-        d f_l(X,B)/d B_k`` evaluated at the i'th data point.  If `q == 1`, then
+        d f_l(X,B)/d B_k`` evaluated at the ith data point.  If `q == 1`, then
         the return array is only rank-2 and with shape `(p, n)`.
     `fjacd`
         as with fjacb, only the return array's shape is `(q, m, n)`
-        such that ``fjacd(x,beta)[l,j,i] = d f_l(X,B)/d X_j`` at the i'th data
+        such that ``fjacd(x,beta)[l,j,i] = d f_l(X,B)/d X_j`` at the ith data
         point.  If `q == 1`, then the return array's shape is `(m, n)`. If
         `m == 1`, the shape is (q, n). If `m == q == 1`, the shape is `(n,)`.
 
@@ -550,7 +550,7 @@ class Output(object):
     beta : ndarray
         Estimated parameter values, of shape (q,).
     sd_beta : ndarray
-        Standard errors of the estimated parameters, of shape (p,).
+        Standard deviations of the estimated parameters, of shape (p,).
     cov_beta : ndarray
         Covariance matrix of the estimated parameters, of shape (p,p).
     delta : ndarray, optional
@@ -585,8 +585,8 @@ class Output(object):
     Notes
     -----
     Takes one argument for initialization, the return value from the
-    function `odr`. The attributes listed as "optional" above are only
-    present if `odr` was run with ``full_output=1``.
+    function `~scipy.odr.odr`. The attributes listed as "optional" above are
+    only present if `~scipy.odr.odr` was run with ``full_output=1``.
 
     """
 
@@ -657,11 +657,13 @@ class ODR(object):
         ODRPACK User's Guide if you absolutely must set the value here. Use the
         method set_iprint post-initialization for a more readable interface.
     errfile : str, optional
-        string with the filename to print ODRPACK errors to. *Do Not Open
-        This File Yourself!*
+        string with the filename to print ODRPACK errors to. If the file already
+        exists, an error will be thrown. The `overwrite` argument can be used to
+        prevent this. *Do Not Open This File Yourself!*
     rptfile : str, optional
-        string with the filename to print ODRPACK summaries to. *Do Not
-        Open This File Yourself!*
+        string with the filename to print ODRPACK summaries to. If the file
+        already exists, an error will be thrown. The `overwrite` argument can be
+        used to prevent this. *Do Not Open This File Yourself!*
     ndigit : int, optional
         integer specifying the number of reliable digits in the computation
         of the function.
@@ -682,7 +684,7 @@ class ODR(object):
     maxit : int, optional
         integer specifying the maximum number of iterations to perform. For
         first runs, maxit is the total number of iterations performed and
-        defaults to 50.  For restarts, maxit is the number of additional
+        defaults to 50. For restarts, maxit is the number of additional
         iterations to perform and defaults to 10.
     stpb : array_like, optional
         sequence (``len(stpb) == len(beta0)``) of relative step sizes to compute
@@ -695,7 +697,7 @@ class ODR(object):
         all observations.
     sclb : array_like, optional
         sequence (``len(stpb) == len(beta0)``) of scaling factors for the
-        parameters.  The purpose of these scaling factors are to scale all of
+        parameters. The purpose of these scaling factors are to scale all of
         the parameters to around unity. Normally appropriate scaling factors
         are computed if this argument is not specified. Specify them yourself
         if the automatic procedure goes awry.
@@ -710,6 +712,9 @@ class ODR(object):
     iwork : ndarray, optional
         array to hold the integer-valued working data for ODRPACK. When
         restarting, takes the value of self.output.iwork.
+    overwrite : bool, optional
+        If it is True, output files defined by `errfile` and `rptfile` are
+        overwritten. The default is False.
 
     Attributes
     ----------
@@ -726,7 +731,8 @@ class ODR(object):
     def __init__(self, data, model, beta0=None, delta0=None, ifixb=None,
         ifixx=None, job=None, iprint=None, errfile=None, rptfile=None,
         ndigit=None, taufac=None, sstol=None, partol=None, maxit=None,
-        stpb=None, stpd=None, sclb=None, scld=None, work=None, iwork=None):
+        stpb=None, stpd=None, sclb=None, scld=None, work=None, iwork=None,
+        overwrite=False):
 
         self.data = data
         self.model = model
@@ -741,6 +747,16 @@ class ODR(object):
         else:
             self.beta0 = _conv(beta0)
 
+        if ifixx is None and data.fix is not None:
+            ifixx = data.fix
+
+        if overwrite:
+            # remove output files for overwriting.
+            if rptfile is not None and os.path.exists(rptfile):
+                os.remove(rptfile)
+            if errfile is not None and os.path.exists(errfile):
+                os.remove(errfile)
+
         self.delta0 = _conv(delta0)
         # These really are 32-bit integers in FORTRAN (gfortran), even on 64-bit
         # platforms.
@@ -1070,7 +1086,7 @@ class ODR(object):
                  'ndigit', 'taufac', 'sstol', 'partol', 'maxit', 'stpb',
                  'stpd', 'sclb', 'scld', 'work', 'iwork']
 
-        if self.delta0 is not None and self.job % 1000 // 10 == 1:
+        if self.delta0 is not None and (self.job // 10000) % 10 == 0:
             # delta0 provided and fit is not a restart
             self._gen_work()
 
